/*
Copyright (C) 2004  The PARI group.

This file is part of the PARI package.

PARI/GP is free software; you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free Software
Foundation. It is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY WHATSOEVER.

Check the License for details. You should have received a copy of it, along
with the package; see the file 'COPYING'. If not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */

#include "pari.h"
#include "tree.h"
#include "anal.h"
const char *Ffuncname[]={"Fseq",
		   "Fmatrix","FmatrixL","FmatrixR",
		   "Faffect",
		   "Ffacteurmat",
		   "Fmatrixelts","Fmatrixlines",
		   "Femptyvec","Femptymat","Fmat",
		   "Flistarg",

		   "Fconst","Fsmall","Fgnil",
		   "Frefarg","Ftag",
		   "Fentry","Fentryfunc","Fdeffunc",

		   /*These nodes are generated by genblock, not by parser*/
		   "Fblock"
};

extern char *gp_function_name;

typedef GEN (*PFGEN)(ANYARG);
#define _ARGS_ argvec[0], argvec[1], argvec[2], argvec[3],\
               argvec[4], argvec[5], argvec[6], argvec[7], argvec[8]

GEN
grandopari(int a)
{
  GEN res;
  long m=1;
  if (tree[a].f!=Fentryfunc || tree[a].x!=OPpow)
    res=evalnode_nobrk(a);
  else
  {
    int y=tree[a].y;
    res=evalnode(tree[y].x);
    m=itos(evalnode_nobrk(tree[y].y));
  }
  return ggrando(res,m); 
}

GEN andpari(int a, int b)
{
  GEN g=evalnode(a);
  if (!g) return NULL;
  if (gcmp0(g))
    return gzero;
  g=evalnode(b);
  if (!g) return NULL;
  return gcmp0(g)?gzero:gun;
}

GEN orpari(int a, int b)
{
  GEN g=evalnode(a);
  if (!g) return NULL;
  if (!gcmp0(g))
    return gun;
  g=evalnode(b);
  if (!g) return NULL;
  return gcmp0(g)?gzero:gun;
}

GEN gmule(GEN *x, GEN y)
{
  *x=gmul(*x,y);
  return *x;
}

GEN gdive(GEN *x, GEN y)
{
  *x=gdiv(*x,y);
  return *x;
}

GEN gdivente(GEN *x, GEN y)
{
  *x=gdivent(*x,y);
  return *x;
}

GEN gdivrounde(GEN *x, GEN y)
{
  *x=gdivround(*x,y);
  return *x;
}

GEN gmode(GEN *x, GEN y)
{
  *x=gmod(*x,y);
  return *x;
}

GEN gshiftle(GEN *x, long n)
{
  *x=gshift(*x,n);
  return *x;
}

GEN gshiftre(GEN *x, long n)
{
  *x=gshift(*x,-n);
  return *x;
}

GEN gadde(GEN *x, GEN y)
{
  *x=gadd(*x,y);
  return *x;
}

GEN gaddone(GEN *x)
{
  *x=typ(*x)==t_INT?addis(*x,1):gaddgs(*x,1);
  return *x;
}

GEN gsube(GEN *x, GEN y)
{
  *x=gsub(*x,y);
  return *x;
}

GEN gsubone(GEN *x)
{
  *x=typ(*x)==t_INT?subis(*x,1):gsubgs(*x,1);
  return *x;
}

GEN gshift_right(GEN x, long n) { return gshift(x,-n); }

extern GEN gnil;

char *current_func_name;

void localpari(GEN v)
{
  int i;
  for(i=1;i<lg(v);i++)
    evalnode_nobrk(v[i]);
}

void globalpari(GEN v)
{
  int i;
  for(i=1;i<lg(v);i++)
    evalnode_nobrk(v[i]);
}

entree *getentry(int n)
{
  if (tree[n].f!=Fentry) 
    err(talker,"identifier required (found [%s])",Ffuncname[tree[n].f]);
  else
  {
    int x=tree[n].x;
    char *name=(char *)value[x].val.str;
    char *p;
    entree *ep; 
    if (current_func_name)
    {
      p=gpmalloc(strlen(current_func_name)+strlen(name)+2);
      sprintf(p,"%s:%s",current_func_name,name);
    }
    else 
      p=name;
    ep = is_entry(p);
    if (!ep)
      ep=fetch_local_var(p,name);
    if (current_func_name)
      free(p);
    return ep;
  }
  return NULL;/*NOT REACHED*/
}

entree *getentryfunc(int n)
{
  if (tree[n].f!=Fentryfunc) 
    err(talker,"function required (found [%s])",Ffuncname[tree[n].f]);
  else
  {
    int x=tree[n].x;
    char *name=(char *)value[x].val.str;
    entree *ep; 
    ep = is_entry(name);
    if (!ep)
      ep=fetch_named_var(name);
    return ep;
  }
  return NULL;/*NOT REACHED*/
}

GEN listtogen(int n, int f)
{
  GEN z;
  int x,i,nb;
  if (n==-1 || n==GNOARG) return cgetg(1,t_VECSMALL);
  for(x=n, i=0; tree[x].f==f ;x=tree[x].x, i++);
  nb=i+1;
  z=cgetg(nb+1,t_VECSMALL);
  for (x=n; i>0; z[i+1]=tree[x].y, x=tree[x].x, i--);
  z[1]=x;
  return z;
}

static GEN
cattovec(int n, int fnum)
{
  int x=n, i=0, nb;
  GEN stack;
  if (n==GNOARG) return cgetg(1,t_VECSMALL);
  while(1)
  {
    int xx=tree[x].x;
    int xy=tree[x].y;
    if (tree[x].f!=Fentryfunc || xx!=fnum) break;
    x=tree[xy].x;
    i++;
  }
  nb=i+1;
  stack=cgetg(nb+1,t_VECSMALL);
  for(x=n;i>0;i--)
  {
    int y=tree[x].y;
    x=tree[y].x;
    stack[i+1]=tree[y].y;
  }
  stack[1]=x;
  return stack;
}

void evallvalue(int n, matcomp *C);

GEN evalfunc(int n)
{
  int i,j;
  int x=tree[n].x;
  char *name=(char *)value[x].val.str;
  int ret;
  GEN res;
  char const *p,*q;
  char c;
  PPproto mod;
  entree *ep=tree[n].e;
  void *call; 
  GEN argvec[9];
  long m;
  GEN arg=tree[n].list;
  int nb;
  int ind_pointer=0;
  gp_pointer ptrs[9];
  if (!arg)
  {
    arg=listtogen(tree[n].y,Flistarg);
    tree[n].list=gclone(arg);
  }
  nb=lg(arg)-1;
  if (!ep)
  {
    ep = is_entry(name);
    if (!ep) 
      err(talker,"unknown function %s",name);
    ep=do_alias(ep);
    tree[n].e=ep;
  }
  if (ep->valence==EpUSER)
  {
    GEN res;
    GEN arg2=listtogen(tree[(int)ep->args].y,Flistarg);
    long nb2=lg(arg2)-1;
    char *old_func=current_func_name;
    for (i=1;i<=nb2;i++)
    {
      int a=arg2[i];
      GEN g=gzero;
      if (i<=nb && arg[i]!=GNOARG)
      {
        if (tree[a].f==Faffect)
          a=tree[a].x;
        g=evalnode_nobrk(arg[i]);
      }
      else
      {
        if (tree[a].f==Faffect)
        {
          g=evalnode_nobrk(tree[a].y);
          a=tree[a].x;
        }
      }
      current_func_name=ep->name;
      if (!tree[a].e) tree[a].e=getentry(a);
      if(!g) return g;
      changevalue(tree[a].e,g);
    }
    res=evaluserfunc((int)ep->value);
    current_func_name=old_func;
    return res;
  }
  p=ep->code; call= ep->value;
  if (!call)
    err(impl,name);
  if      (*p < 'a') { ret = RET_GEN; }
  else if (*p == 'v') { ret = RET_VOID; p++; }
  else if (*p == 'i') { ret = RET_INT;  p++; }
  else if (*p == 'l') { ret = RET_LONG; p++; }
  else ret = RET_GEN;
  i=0; j=1;
  while(*p=='G' && j<=nb)
  {
    argvec[i++]=evalnode_nobrk(arg[j++]);
    p++;
  }
  if (*p)
  {
    q=p; 
    while((mod=parseproto(&p,&c))!=PPend)
    {
      if (j<=nb && arg[j]!=GNOARG && (mod==PPdefault || mod==PPdefaultmulti))
        mod=PPstd;
      switch(mod)
      {
      case PPstd:
        if (j>nb)
          err(talker,"missing mandatory argument for %s",name);
        switch(c)
        {
        case 'G':
          argvec[i]=evalnode_nobrk(arg[j++]);
          break;
        case 'M':
        case 'L':
          argvec[i]=(GEN)itos(evalnode_nobrk(arg[j++]));
          break;
        case 'n':
          argvec[i]=(GEN)varn(evalnode_nobrk(arg[j++]));
          break;
        case '&':
        {
          int a=arg[j++];
          gp_pointer *g = &ptrs[ind_pointer++];
          if (tree[a].f==Frefarg)
            a=tree[a].x;
          if (tree[a].f==Fentry)
          {
            if (!tree[a].e) tree[a].e=getentry(a);
            g->ep = tree[a].e;
            g->x = (GEN) g->ep->value;
          }
          else
          {
            evallvalue(a, &(g->c));
            g->ep = NULL;
            g->x = *(g->c.ptcell);
          }
          argvec[i] = (GEN)&(g->x); break;
          break;
        }
        case 'I':
        case 'E':
          argvec[i]=(GEN)arg[j++];
          break;
        case 'S':
        case 'V':
          if (!tree[arg[j]].e) tree[arg[j]].e=getentry(arg[j]);
          argvec[i]=(GEN)tree[arg[j++]].e;
          break;
        case 's':
          argvec[i]=(GEN)GENtostr(evalnode_nobrk(arg[j++]));
          break;
        case 'r':
          {
            int a=arg[j++];
            if (tree[a].f==Fentry)
              argvec[i]=(GEN)value[tree[a].x].val.str;
            else
              argvec[i]=(GEN)GENtostr(evalnode_nobrk(a));
          }
          break;
        default:
          err(talker,"Unknown prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPauto:
        switch(c)
        {
        case 'p':
          argvec[i]=(GEN)prec;
          break;
        case 'P':
          argvec[i]=(GEN)precdl;
          break;
        }
        break;
      case PPdefault:
        j++;
        switch(c)
        {
        case 'G':
        case '&':
        case 'V':
        case 'r':
          argvec[i]=(GEN)NULL;
          break;
        case 'n':
          argvec[i]=(GEN)-1;
          break;
        case 'E':
        case 'I':
          argvec[i]=(GEN)-1;
          break;
        default:
          err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPdefaultmulti:
        j++;
        switch(c)
        {
        case 'L':
        case 'M':
          argvec[i]=(GEN)strtol(q+1,NULL,10);
          break;
        case 's':
        case 'r':
          argvec[i]=(GEN)"";
          break;
        default:
          err(talker,"Unknown default in prototype code `%c' for `%s'",c,name);
        }
        break;
      case PPstar:
        switch(c)
        {
        case 'V':
          {
            int x=tree[arg[j]].x;
            int y=tree[arg[j]].y;
            if (tree[arg[j]].f!=Faffect)
              err(talker,"= expected");
            if (!tree[x].e) tree[x].e=getentry(x);
            argvec[i++]=(GEN)tree[x].e;
            argvec[i]=(GEN)evalnode_nobrk(y);
            j++;
          }
          break;
        case 'I':
          argvec[i]=(GEN)arg;
          j=nb+1;
          break;
        case 's':
          {
            int n=nb+1-j;
            int k,l,l1,m;
            GEN r,g=cgetg(n+1,t_VEC);
            for(l1=0,k=1;k<=n;k++)
            {
              g[k]=(long)cattovec(arg[j+k-1],OPcat);
              l1+=lg(g[k])-1;
            }
            r=cgetg(l1+1,t_VEC);
            for(m=1,k=1;k<=n;k++)
            {
              for(l=1;l<lg(g[k]);l++)
                r[m++]=(long)evalnode_nobrk(mael(g,k,l));
            }
            j=nb+1;
            argvec[i]=r;
          }
          break;
        default:
          err(talker,"Unknown prototype code `%c*' for `%s'",c,name);
        }
        break;
      default:
        err(bugparier,"PPproto %d in gencallfunc",mod);
      }
      i++;
      q=p;
    }
  }
  if (j<=nb)
    err(talker,"too many arguments for %s",name);
  for (  ;i<9;i++) argvec[i]=NULL;
  gp_function_name=name;
  switch (ret)
  {
  case RET_GEN: 
    res = ((PFGEN)call)(_ARGS_);
    break;

  case RET_INT:
    m = (long)((int (*)(ANYARG))call)(_ARGS_);
    res = stoi(m); break;

  case RET_LONG:
    m = ((long (*)(ANYARG))call)(_ARGS_);
    res = stoi(m); break;

  case RET_VOID:
    ((void (*)(ANYARG))call)(_ARGS_);
    res = gnil; break;
  default:
    res = gnil;
  }
  gp_function_name=NULL;
  if (ind_pointer) check_pointers(ptrs, ind_pointer);
  return evalparifunc(res);
}

static long
check_array_index(long c, long max)
{
  if (c < 1 || c >= max)
  {
    char s[80];
    sprintf(s,"array index (%ld) out of allowed range ",c);
    if (max == 1) strcat(s, "[none]");
    else if (max == 2) strcat(s, "[1]");
    else sprintf(s,"%s[1-%ld]",s,max-1);
    err(talker,s);
  }
  return c;
}

GEN 
evalfacteurmat(int n)
{ 
  int x=tree[n].x;
  int y=tree[n].y;
  int yx=tree[y].x;
  int yy=tree[y].y;
  GEN p=evalnode(x),p2;
  long t=typ(p);
  long b,c;
  long f=tree[y].f;
  if (f==Fmatrix && yy==-1)
    switch(t)
    {
    case t_VEC: case t_COL:
      c = itos(evalnode(yx));
      check_array_index(c, lg(p));
      return (GEN) p[c];
    case t_LIST:
      c = itos(evalnode(yx));
      check_array_index(c,lgeflist(p)-1);
      return (GEN) p[c+1]; 
    case t_VECSMALL:
      c = itos(evalnode(yx));
      check_array_index(c,lg(p));
      return stoi(p[c]);
    case t_MAT: 
      err(talker,"matrix has no simple components");
      return NULL; /*Not reached*/
    default:
      err(talker,"object has no components");
      return NULL; /*Not reached*/
    }
  if (t!=t_MAT)
    err(talker,"object has no double components");
  switch(f)
  {
  case Fmatrix:
    c=itos(evalnode(yx));
    check_array_index(c,lg(p));
    b=itos(evalnode(yy));
    check_array_index(b,lg(p[c]));
    return gcoeff(p,b,c);
  case FmatrixR:
    c=itos(evalnode(yx));
    check_array_index(c,lg(p));
    return (GEN) p[c];
  case FmatrixL:
    b=itos(evalnode(yx));
    check_array_index(b,lg(p[1]));
    p2 = cgetg(lg(p),t_VEC);
    for (c=1; c<lg(p); c++) p2[c] = coeff(p,b,c);
    return p2;
  }
  return NULL;
}

void
evallvalue_rec(int n, matcomp *C)
{ 
  int x=tree[n].x;
  int y=tree[n].y;
  int yx=tree[y].x;
  int yy=tree[y].y;
  GEN p,p2;
  long t;
  long b,c;
  long f=tree[y].f;
  
  evallvalue(x,C);
  p=*C->ptcell;
  t=typ(p);
  C->full_col = C->full_row = 0;
  if (f==Fmatrix && yy==-1)
    switch(t)
    {
    case t_VEC: case t_COL:
      c = itos(evalnode(yx));
      check_array_index(c, lg(p));
      C->ptcell = (GEN *) p+c;
      break;
    case t_LIST:
      c = itos(evalnode(yx));
      check_array_index(c,lgeflist(p)-1);
      C->ptcell = (GEN *) p+c+1;
      break;
    case t_VECSMALL:
      c = itos(evalnode(yx));
      check_array_index(c,lg(p));
      C->ptcell = (GEN *) p+c;
      break;
    case t_MAT: 
      err(talker,"matrix has no simple components");
    default:
      err(talker,"object has no components");
    }
  else
  {
    if (t!=t_MAT)
      err(talker,"object has no double components");
    switch(f)
    {
    case Fmatrix:
      c=itos(evalnode(yx));
      check_array_index(c,lg(p));
      b=itos(evalnode(yy));
      check_array_index(b,lg(p[c]));
      C->ptcell = (GEN *) gel(p,c)+b;
      break;
    case FmatrixR:
      c=itos(evalnode(yx));
      check_array_index(c,lg(p));
      C->ptcell = (GEN *) p+c;
      C->full_col = 1; 
      break;
    case FmatrixL:
      b=itos(evalnode(yx));
      check_array_index(b,lg(p[1]));
      p2 = cgetg(lg(p),t_VEC);
      for (c=1; c<lg(p); c++) p2[c] = coeff(p,b,c);
      C->ptcell = (GEN *) &p2;
      C->full_row = 1;
      break;
    }
  }
}

void
evallvalue(int n, matcomp *C)
{
  if (tree[n].f!=Fentry)
  {
    evallvalue_rec(n,C); 
    return;
  }
  if (!tree[n].e) tree[n].e=getentry(n); 
  C->full_col = C->full_row = 0;
  C->parent = (GEN)    tree[n].e->value;
  C->ptcell = (GEN *) &tree[n].e->value;
}
 
GEN evalvector(int n)
{
  int x=tree[n].x;
  int i;
  GEN ret;
  GEN arg=listtogen(x,Fmatrixelts);
  long l=lg(arg);
  ret=cgetg(l,t_VEC); 
  for(i=1;i<l;i++)
    ret[i]=(long)evalnode(arg[i]);
  return ret; 
}
GEN evalmatrix(int n)
{
  int x=tree[n].x;
  int i,j;
  GEN ret;
  GEN line=listtogen(x,Fmatrixlines);
  long nbline = lg(line), nbcol=0;
  GEN lines=cgetg(nbline,t_MAT);
  for(i=1;i<nbline;i++)
  {
    long k;
    lines[i]=(long)listtogen(line[i],Fmatrixelts);
    k=lg(lines[i]);
    if (i>1 && k!=nbcol)
      err(talker,"Matrix must be rectangular");
    nbcol=k;
  }
  ret=cgetg(nbcol,t_MAT);
  for(i=1;i<nbcol;i++)
  {
    GEN v=cgetg(nbline,t_COL);
    for(j=1; j<nbline; j++)
      v[j]=(long) evalnode(mael(lines,j,i));
    ret[i]=(long) v;
  }
  return ret;
}

GEN evalnode(int n)
{
  int x,y;
  if (n<0)
    err(talker,"missing mandatory argument");
  x=tree[n].x;
  y=tree[n].y;
  if (DEBUGLEVEL>4)
    fprintferr("[%d]:%s(%d,%d)",n,Ffuncname[tree[n].f],x,y);
   
  switch(tree[n].f)
  {
  case Fseq:
    return evalnode(x)?evalnode(y):NULL;
  case Fmatrix:
  case FmatrixL:
  case FmatrixR:
    break;
  case Ffacteurmat:
    return evalfacteurmat(n);
  case Faffect:
    if (tree[x].f==Fentry)
    {
      GEN g=evalnode(y);
      if (!g) return g;
      if (!tree[x].e) tree[x].e=getentry(x); 
      changevalue(tree[x].e,g);
      return g;
    }
    else
    {
      pari_sp av=avma;
      matcomp C;
      GEN g;
      evallvalue(x,&C);
      g=evalnode(y);
      if (!g) return g;
      return change_compo(av,&C, g);
    }
  case Fconst:
    switch(value[x].type)
    {
    case CSTsmall:
      return stoi(value[x].val.small);
    case CSTsmallreal:
      return dbltor(value[x].val.small*1.);
    case CSTgen: 
      return value[x].val.gen;
    case CSTdot: 
      return itor(value[x].val.gen,prec);
    case CSTstr:
      return strtoGENstr((char *)value[x].val.str);
    case CSTquote:
      return polx[fetch_user_var((char *)value[x].val.str)];
    }
    break;
  case Fsmall:
    return stoi(x);
  case Fmatrixelts:
  case Fmatrixlines:
    break;
  case Femptyvec:
    return cgetg(1,t_VEC);
  case Femptymat:
    return cgetg(1,t_MAT);
  case Fmat:
    if (tree[x].f==Fmatrixlines)
      /*if there is at least 2 lines it is a matrix*/
      return evalmatrix(n);
    else
      /*else it is a vector*/
      return evalvector(n);
  case Flistarg:
    break;
  case Frefarg:
    err(talker,"unexpected &"); 
    break;
  case Fentry:
    if (!tree[n].e)
    {
      char *name=(char *)value[x].val.str;
      entree *ep=is_entry(name);
      if (ep && EpVALENCE(ep)!=EpVAR && EpVALENCE(ep)!=EpGVAR)
      {
        tree[n].f=Fentryfunc;
        tree[n].e=ep;
        return evalfunc(n);
      }
      tree[n].e=getentry(n);
    }
    return gcopy(tree[n].e->value);
  case Fentryfunc:
    return evalfunc(n);
  case Fdeffunc:
    {
      entree *ep = getentryfunc(x);
      ep->valence=EpUSER;
      ep->value = (void *) y;
      ep->args = (void *) x;
      return gnil;
    }
  case Fblock:
    break;
  case Fgnil:
    return gnil;
  case Ftag:
    return evalnode(x);    
  }
  err(talker,"node %s not implemented yet\n",Ffuncname[tree[n].f]);
  return gnil;
}


